@@ -4,9 +4,9 @@ import math
 import os
 
 import dill
-from efficientnet.tfkeras import EfficientNetB0, EfficientNetB3, EfficientNetB5
 import skopt
 import tensorflow as tf
+from efficientnet.tfkeras import EfficientNetB0, EfficientNetB3, EfficientNetB5, EfficientNetB7
 
 from decavision.utils import training_utils
 from decavision.utils import utils
@@ -25,20 +25,23 @@ class ImageClassifier:
             folders train and val, filenames of the form filenumber-numberofimages.tfrec
         batch_size (int): size of batches of data used for training
         transfer_model (str): pretrained model to use for transfer learning, can be one of Inception,
-            Xception, Inception_Resnet, Resnet, (EfficientNet) B0, B3 or B5
+            Xception, Inception_Resnet, Resnet, (EfficientNet) B0, B3, B5 or B7
+        augment (boolean): Whether to augment the training data, default is True
     """
 
-    def __init__(self, tfrecords_folder, batch_size=128, transfer_model='Inception'):
+    def __init__(self, tfrecords_folder, batch_size=128, transfer_model='Inception', augment=True):
 
         self.tfrecords_folder = tfrecords_folder
         self.use_TPU, self.use_GPU = utils.check_PU()
         if self.use_TPU and batch_size % 8:
-            print('Batch size {} is not multiple of 8, required for TPU'.format(batch_size))
-            batch_size = 8 * round(batch_size/8)
+            print(
+                'Batch size {} is not multiple of 8, required for TPU'.format(batch_size))
+            batch_size = 8 * round(batch_size / 8)
             print('New batch size is {}'.format(batch_size))
         self.batch_size = batch_size
         self.transfer_model = transfer_model
-        
+        self.augment = augment
+
         # We expect the classes to be saved in the same folder where tfrecords are
         self.categories = utils.load_classes(tfrecords_folder)
         print('Classes ({}) :'.format(len(self.categories)))
@@ -48,42 +51,46 @@ class ImageClassifier:
             os.path.join(tfrecords_folder, 'train'))
         self.nb_train_shards = len(train_tfrecords)
         print('Training tfrecords = {}'.format(self.nb_train_shards))
-        
+
         val_tfrecords = tf.io.gfile.listdir(
             os.path.join(tfrecords_folder, 'val'))
         self.nb_val_shards = len(val_tfrecords)
         print('Val tfrecords = {}'.format(self.nb_val_shards))
-        
+
         # Expected tfrecord file name : filenumber-numberofimages.tfrec (02-2223.tfrec)
         self.nb_train_images = 0
         for train_tfrecord in train_tfrecords:
-            self.nb_train_images += int(train_tfrecord.split('.')[0].split('-')[1])
-        print('Training images = '+str(self.nb_train_images))
-        
+            self.nb_train_images += int(train_tfrecord.split('.')
+                                        [0].split('-')[1])
+        print('Training images = ' + str(self.nb_train_images))
+
         nb_val_images = 0
         for val_tfrecord in val_tfrecords:
             nb_val_images += int(val_tfrecord.split('.')[0].split('-')[1])
-        print('Val images = '+str(nb_val_images))
-        
-        self.training_shard_size = math.ceil(self.nb_train_images/self.nb_train_shards)
+        print('Val images = ' + str(nb_val_images))
+
+        self.training_shard_size = math.ceil(
+            self.nb_train_images / self.nb_train_shards)
         print('Training shard size = {}'.format(self.training_shard_size))
 
-        val_shard_size = math.ceil(nb_val_images/self.nb_val_shards)
+        val_shard_size = math.ceil(nb_val_images / self.nb_val_shards)
         print('Val shard size = {}'.format(val_shard_size))
 
-        print('Training batch size = '+str(self.batch_size))
+        print('Training batch size = ' + str(self.batch_size))
         self.steps_per_epoch = int(self.nb_train_images / self.batch_size)
-        print('Training steps per epochs = '+str(self.steps_per_epoch))
+        print('Training steps per epochs = ' + str(self.steps_per_epoch))
 
-        print('Val batch size = '+str(self.batch_size))
+        print('Val batch size = ' + str(self.batch_size))
         self.validation_steps = int(nb_val_images / self.batch_size)
-        print('Val steps per epochs = '+str(self.validation_steps))
+        print('Val steps per epochs = ' + str(self.validation_steps))
 
-        if transfer_model in ['Inception', 'Xception', 'Inception_Resnet', 'B3', 'B5']:
+        if transfer_model in ['Inception', 'Xception', 'Inception_Resnet', 'B3', 'B5', 'B7']:
             self.target_size = (299, 299)
         else:
             self.target_size = (224, 224)
 
+        print("Data augmentation during training: " + str(augment))
+
     def _get_dataset(self, is_training, nb_readers):
         """
         Extract data from tfrecords into a tf.data iterator. If data is used for training, it is shuffled each time
@@ -98,15 +105,15 @@ class ImageClassifier:
         Returns:
             tf.data.dataset: iterable dataset with content of relevant tfrecords (images and labels)
         """
-        
+
         def _read_tfrecord(example):
             """ Extract image and label from single tfrecords example."""
             features = {
-                    'image': tf.io.FixedLenFeature((), tf.string),
-                    'label': tf.io.FixedLenFeature((), tf.int64),
-                }
+                'image': tf.io.FixedLenFeature((), tf.string),
+                'label': tf.io.FixedLenFeature((), tf.int64),
+            }
             example = tf.io.parse_single_example(example, features)
-                
+
             image = tf.image.decode_jpeg(example['image'], channels=3)
             image = tf.image.convert_image_dtype(image, dtype=tf.float32)
             feature = tf.image.resize(image, [*self.target_size])
@@ -119,32 +126,64 @@ class ImageClassifier:
             dataset = tf.data.TFRecordDataset(
                 filenames, buffer_size=buffer_size)
             return dataset
-    
-        file_pattern = os.path.join(self.tfrecords_folder, "train/*" if is_training else "val/*")
+
+        file_pattern = os.path.join(
+            self.tfrecords_folder, "train/*" if is_training else "val/*")
         dataset = tf.data.Dataset.list_files(file_pattern, shuffle=is_training)
 
         # Enable non-determinism only for training.
         options = tf.data.Options()
         options.experimental_deterministic = not is_training
         dataset = dataset.with_options(options)
-        dataset = dataset.interleave(_load_dataset, nb_readers, num_parallel_calls=AUTO)
+        dataset = dataset.interleave(
+            _load_dataset, nb_readers, num_parallel_calls=AUTO)
         if is_training:
             # Shuffle only for training.
-            dataset = dataset.shuffle(buffer_size=math.ceil(self.training_shard_size*self.nb_train_shards/4))
+            dataset = dataset.shuffle(buffer_size=math.ceil(
+                self.training_shard_size * self.nb_train_shards / 4))
         dataset = dataset.repeat()
         dataset = dataset.map(_read_tfrecord, num_parallel_calls=AUTO)
-        dataset = dataset.batch(batch_size=self.batch_size, drop_remainder=self.use_TPU)
+        dataset = dataset.batch(
+            batch_size=self.batch_size, drop_remainder=self.use_TPU)
         dataset = dataset.prefetch(AUTO)
         return dataset
 
+    def data_augment(self, image, label):
+        """
+        Data augmentation pipeline which augments the data by randomly flipping, changing brightness
+        and saturation for each batch during training a model.
+
+        References:
+        https://www.wouterbulten.nl/blog/tech/data-augmentation-using-tensorflow-data-dataset/#code
+        https://www.tensorflow.org/tutorials/images/data_augmentation
+
+        Returns:
+            augmented images and labels
+        """
+        # Flip
+        image = tf.image.random_flip_left_right(image)
+        # Brightness and saturation
+        image = tf.image.random_brightness(image, max_delta=32.0 / 255.0)
+        image = tf.image.random_saturation(image, lower=0.5, upper=1.5)
+        # Make sure the image is still in [0, 1] range, after augmentation
+        image = tf.clip_by_value(image, 0.0, 1.0)
+        return image, label
+
     def get_training_dataset(self):
         """
-        Extract data from training tfrecords located in tfrecords_folder. Data is shuffled.
+        Extract data from training tfrecords located in tfrecords_folder. Data is shuffled and augmented.
 
         Returns:
             tf.data.dataset: iterable dataset with content of training tfrecords (images and labels)
         """
-        return self._get_dataset(True, self.nb_train_shards)
+        if self.augment is True:
+            dataset = self._get_dataset(True, self.nb_train_shards)
+            # Augment data
+            dataset = dataset.map(self.data_augment, num_parallel_calls=AUTO)
+            return dataset
+        else:
+            dataset = self._get_dataset(True, self.nb_train_shards)
+        return dataset
 
     def get_validation_dataset(self):
         """ Extract data from validation tfrecords located in tfrecords_folder.
@@ -197,6 +236,11 @@ class ImageClassifier:
             base_model = EfficientNetB5(weights='imagenet', include_top=False,
                                         input_shape=(*self.target_size, 3))
             base_model_last_block = 417  # last block 559, two blocks 417
+
+        elif self.transfer_model == 'B7':
+            base_model = EfficientNetB7(weights='imagenet', include_top=False,
+                                        input_shape=(*self.target_size, 3))
+            base_model_last_block = None  # all layers trainable
         else:
             base_model = tf.keras.applications.InceptionV3(weights='imagenet',
                                                            include_top=False, input_shape=(*self.target_size, 3))
@@ -217,18 +261,21 @@ class ImageClassifier:
                                       kernel_regularizer=tf.keras.regularizers.l2(l=l2_lambda))(x)
             # scale: When the next layer is linear (also e.g. nn.relu), this can be disabled since the
             # scaling can be done by the next layer.
-            x = tf.keras.layers.BatchNormalization(scale=activation != 'relu')(x)
+            x = tf.keras.layers.BatchNormalization(
+                scale=activation != 'relu')(x)
             x = tf.keras.layers.Activation(activation=activation)(x)
             x = tf.keras.layers.Dropout(rate=dropout)(x)
 
-        x = tf.keras.layers.Dense(len(self.categories), name='logs')(x)  # Output layer
-        predictions = tf.keras.layers.Activation('softmax', name='preds')(x)  # Output activation
+        x = tf.keras.layers.Dense(
+            len(self.categories), name='logs')(x)  # Output layer
+        predictions = tf.keras.layers.Activation(
+            'softmax', name='preds')(x)  # Output activation
         loss = 'sparse_categorical_crossentropy'
         metrics = ['sparse_categorical_accuracy']
 
         return tf.keras.Model(inputs=base_model.input, outputs=predictions), base_model_last_block, loss, metrics
 
-    def fit(self, save_model=None, export_model=None, min_accuracy=None,
+    def fit(self, save_model=None, export_model=None, patience=0,
             epochs=5, hidden_size=1024, learning_rate=1e-3, learning_rate_fine_tuning=1e-4,
             dropout=0.5, l2_lambda=5e-4, fine_tuning=True,
             verbose=True, logs=None, activation='swish'):
@@ -251,44 +298,54 @@ class ImageClassifier:
             verbose (bool): show details of training or not
             fine_tuning (bool): fine tune pretrained model or not
             l2_lambda (float): amount of L2 regularization to include in extra layer
-            min_accuracy (float): if specified, stop training when improvement in val accuracy is smaller than min_accuracy
+            patience (int): if non zero, stop training when improvement in val accuracy is not observed for the given number of epochs. If used, best model is restored when training is stopped
             logs (str): if specified, tensorboard is used and logs are saved at this location
         """
-        callbacks = None
+
+        # use reduce learning rate and early stopping callbacks
+        reduce_lr = tf.keras.callbacks.ReduceLROnPlateau(monitor='val_sparse_categorical_accuracy',
+                                                         factor=0.1,
+                                                         patience=5,
+                                                         mode='max')
+        callbacks = [reduce_lr]
+
         if logs:
-            logdir = os.path.join(logs, datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
+            logdir = os.path.join(
+                logs, datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
             print('Fit log dir : ' + logdir)
             tensorboard_callback = tf.keras.callbacks.TensorBoard(logdir)
-            callbacks = [tensorboard_callback]
+            callbacks.append(tensorboard_callback)
 
-        # if we want to stop training when no sufficient improvement in accuracy has been achieved
-        if min_accuracy:
+        # if we want to stop training when no sufficient improvement in validation metric has been achieved
+        if patience:
             early_stop = tf.keras.callbacks.EarlyStopping(monitor='val_sparse_categorical_accuracy',
-                                                          min_delta=min_accuracy)
-            if not callbacks:
-                callbacks = [early_stop]
-            else:
-                callbacks.append(early_stop)
-        
+                                                          patience=patience,
+                                                          restore_best_weights=True)
+            callbacks.append(early_stop)
+
         # compile the model and fit the model
         if self.use_TPU:
             tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver()
             tf.config.experimental_connect_to_cluster(tpu_cluster_resolver)
             tf.tpu.experimental.initialize_tpu_system(tpu_cluster_resolver)
-            strategy = tf.distribute.experimental.TPUStrategy(tpu_cluster_resolver)
-            
+            strategy = tf.distribute.experimental.TPUStrategy(
+                tpu_cluster_resolver)
+
             with strategy.scope():
-                model, base_model_last_block, loss, metrics = self._create_model(activation, hidden_size, dropout, l2_lambda)
+                model, base_model_last_block, loss, metrics = self._create_model(
+                    activation, hidden_size, dropout, l2_lambda)
                 print('Compiling for TPU')
                 optimizer = tf.keras.optimizers.Adam(lr=learning_rate)
                 model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
-                    
+
         else:
-            model, base_model_last_block, loss, metrics = self._create_model(activation, hidden_size, dropout, l2_lambda)
-            print('Compiling for GPU') if self.use_GPU else print('Compiling for CPU')
+            model, base_model_last_block, loss, metrics = self._create_model(
+                activation, hidden_size, dropout, l2_lambda)
+            print('Compiling for GPU') if self.use_GPU else print(
+                'Compiling for CPU')
             optimizer = tf.keras.optimizers.Adam(lr=learning_rate)
             model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
-                
+
         print('Fitting')
         history = model.fit(self.get_training_dataset(), steps_per_epoch=self.steps_per_epoch, epochs=epochs,
                             validation_data=self.get_validation_dataset(), validation_steps=self.validation_steps,
@@ -299,10 +356,10 @@ class ImageClassifier:
             print('===========')
             print('Fine-tuning')
             print('===========')
-            
+
             fine_tune_epochs = epochs
             total_epochs = epochs + fine_tune_epochs
-            
+
             print('Unfreezing last block of layers from the base model')
             for layer in model.layers[:base_model_last_block]:
                 layer.trainable = False
@@ -316,7 +373,7 @@ class ImageClassifier:
             print('Recompiling model')
             optimizer = tf.keras.optimizers.Adam(lr=learning_rate_fine_tuning)
             model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
-                
+
             print('Fine tunning')
             history = model.fit(self.get_training_dataset(), steps_per_epoch=self.steps_per_epoch, epochs=total_epochs,
                                 validation_data=self.get_validation_dataset(), validation_steps=self.validation_steps,
@@ -335,7 +392,7 @@ class ImageClassifier:
             self.model.save(export_model)
             print('Model exported')
 
-    def hyperparameter_optimization(self, num_iterations=20, n_random_starts=10, min_accuracy=None, save_results=False):
+    def hyperparameter_optimization(self, num_iterations=20, n_random_starts=10, patience=0, save_results=False):
         """
         Try different combinations of hyperparameters to find the best model possible. Start by trying random
         combinations and after some time learn from th previous tries. Scikit-optimize checkoint is saved
@@ -348,8 +405,7 @@ class ImageClassifier:
             n_random_starts (int): number of random combinations of hyperparameters first tried
             num_iterations (int): total number of hyperparameter combinations to try (aim for a 1:1 to 2:1 ratio
                 num_iterations/n_random_starts)
-            min_accuracy (float): if specified, training in single try will stop when improvement in accuracy
-                is smaller than min_accuracy
+            patience (int): if non zero, stop training when improvement in val accuracy is not observed for the given number of epochs. If used, best model is restored when training is stopped
             save_results (bool): decide to save optimal hyperparameters in hyperparameters_dimensions.pickle when done
         """
         # initialize logging
@@ -363,7 +419,8 @@ class ImageClassifier:
         )
         # declare the hyperparameters search space
         dim_epochs = skopt.space.Integer(low=1, high=6, name='epochs')
-        dim_hidden_size = skopt.space.Integer(low=512, high=2048, name='hidden_size')
+        dim_hidden_size = skopt.space.Integer(
+            low=512, high=2048, name='hidden_size')
         dim_learning_rate = skopt.space.Real(low=1e-6, high=1e-2, prior='log-uniform',
                                              name='learning_rate')
         dim_learning_rate_fine_tuning = skopt.space.Real(low=1e-6, high=1e-2, prior='log-uniform',
@@ -394,9 +451,10 @@ class ImageClassifier:
             default_parameters = [2, 1024, 5e-4, 6e-4, 0.9, 1e-3, True]
             start_from_checkpoint = False
 
-
-        checkpoint_saver = skopt.callbacks.CheckpointSaver('checkpoint.pkl', store_objective=False)
-        checkpoint_downloader = training_utils.CheckpointDownloader('checkpoint.pkl')
+        checkpoint_saver = skopt.callbacks.CheckpointSaver(
+            'checkpoint.pkl', store_objective=False)
+        checkpoint_downloader = training_utils.CheckpointDownloader(
+            'checkpoint.pkl')
         verbose = skopt.callbacks.VerboseCallback(n_total=num_iterations)
 
         @skopt.utils.use_named_args(dimensions=dimensions)
@@ -410,7 +468,8 @@ class ImageClassifier:
             logging.info(f'epochs:{epochs}')
             logging.info(f'hidden_size:{hidden_size}')
             logging.info(f'learning rate:{learning_rate}')
-            logging.info(f'learning rate fine tuning:{learning_rate_fine_tuning}')
+            logging.info(
+                f'learning rate fine tuning:{learning_rate_fine_tuning}')
             logging.info(f'dropout:{dropout}')
             logging.info(f'l2_lambda:{l2_lambda}')
             logging.info(f'fine_tuning:{fine_tuning}')
@@ -419,7 +478,7 @@ class ImageClassifier:
             self.fit(epochs=epochs, hidden_size=hidden_size, learning_rate=learning_rate,
                      learning_rate_fine_tuning=learning_rate_fine_tuning,
                      dropout=dropout, l2_lambda=l2_lambda, fine_tuning=fine_tuning,
-                     min_accuracy=min_accuracy)
+                     patience=patience)
 
             # extract fitness
             fitness = self.fitness
@@ -463,7 +522,8 @@ class ImageClassifier:
         # build results dictionary
         results_dict = {dimensions[i].name: search_result.x[i]
                         for i in range(len(dimensions))}
-        logging.info('Optimal fitness value of:{}'.format(-float(search_result.fun)))
+        logging.info(
+            'Optimal fitness value of:{}'.format(-float(search_result.fun)))
         logging.info('Optimal hyperparameters:{}'.format(results_dict))
 
 
